from fastapi import APIRouter, Depends

from ambient_edge_server.services.authorization_service import AuthorizationService
from ambient_edge_server.services.service_manager import svc_manager

router = APIRouter(prefix="/auth", tags=["auth"])


@router.post("/")
async def authorize_node(
    node_id: str,
    refresh_token: str,
    auth_svc: AuthorizationService = Depends(svc_manager.get_authorization_service),
):
    """Authorize a node to connect to the backend API

    Args:
        node_id (str): Node ID
        refresh_token (str): Refresh token
        auth_svc (AuthorizationService, optional): Defaults to
            Depends(svc_manager.get_authorization_service).

    Returns:
        dict: Response
    """
    await auth_svc.authorize_node(node_id, refresh_token)
    return {"status": "success"}


@router.get("/status")
async def get_auth_status(
    auth_svc: AuthorizationService = Depends(svc_manager.get_authorization_service),
):
    """Get authorization status

    Args:
        auth_svc (AuthorizationService, optional): Defaults to
            Depends(svc_manager.get_authorization_service).

    Returns:
        dict: Response
    """
    result = await auth_svc.verify_authorization_status()
    if result.is_ok():
        return {"status": result.unwrap()}
    return {"status": result.unwrap_err()}



from abc import ABC, abstractmethod
from pathlib import Path

from jinja2 import Environment, FileSystemLoader, TemplateNotFound
from result import Err, Ok, Result

from ambient_client_common.utils import logger
from ambient_edge_server.config import settings
from ambient_edge_server.models.system_service import (
    ServiceConfigBase,
    ServiceConfigLinux,
)


class SystemDaemonRepo(ABC):
    @abstractmethod
    async def render_template_and_save(
        self, service_config: ServiceConfigBase, output_path: Path
    ) -> Result[str, str]:
        """Render the service template and save it to the output path."""

    @abstractmethod
    async def render_template_dry_run(
        self, service_config: ServiceConfigBase
    ) -> Result[str, str]:
        """Render the service template without saving it to disk."""

    @abstractmethod
    async def _render_service_template(
        self, service_config: ServiceConfigBase
    ) -> Result[str, str]:
        """Render the service template."""


class LinuxSystemDaemonRepo(SystemDaemonRepo):
    def __init__(self):
        package_location = Path(__file__).parent.parent
        logger.debug("app.startup() - __file__: {}", __file__)
        logger.debug("app.startup() - package location: {}", package_location)
        self.template_path = package_location / settings.service_template_location
        if not self.template_path.exists():
            logger.warning("Template path does not exist: {}", self.template_path)
        logger.debug("template: {}", self.template_path.absolute())

    async def _render_service_template(
        self, service_config: ServiceConfigLinux
    ) -> Result[str, str]:
        logger.debug("Rendering service template")
        try:
            loader_path = self.template_path.absolute().parent
            logger.debug("Template path: {}", loader_path)
            loader = FileSystemLoader(loader_path)
            logger.debug("Loader: {}", loader.list_templates())
            env = Environment(loader=loader, autoescape=True)
            logger.debug("Environment: {}", env)
            try:
                template = env.get_template("ambient_edge_server.service.jinja2")
            except TemplateNotFound as e:
                logger.error("Template not found: {}", e)
                return Err(str(e))
            logger.debug("Template: {}", template)

            logger.debug(
                "LinuxSystemDaemonRepo._render_service_template - service: {}",
                service_config.model_dump_json(indent=4),
            )
            rendered_content = template.render(service=service_config.model_dump())
            logger.debug(
                "LinuxSystemDaemonRepo._render_service_template - Rendered content: {}",
                rendered_content,
            )
            return Ok(rendered_content)
        except Exception as e:
            logger.error(
                "Failed to render service template: {}", e.with_traceback(None)
            )
            return Err(str(e))

    async def render_template_and_save(
        self,
        service_config: ServiceConfigLinux,
        output_path: Path = Path("/etc/systemd/system/ambient_edge_server.service"),
    ) -> Result[str, str]:
        logger.info(
            "Rendering service template and saving to {}", output_path.absolute()
        )
        try:
            rendered_content = await self._render_service_template(service_config)
            if rendered_content.is_err():
                return rendered_content
            logger.debug("Rendered content: {}", rendered_content)
            output_path.write_text(rendered_content.unwrap())
        except Exception as e:
            logger.error("Failed to render service template: {}", e)
            return Err(str(e))
        finally:
            logger.info("Template saved to {}", output_path.absolute())
            return Ok(f"Template saved to {output_path.absolute()}")

    async def render_template_dry_run(
        self, service_config: ServiceConfigLinux
    ) -> Result[str, str]:
        logger.debug("Rendering service template in dry run mode")
        return await self._render_service_template(service_config)



import datetime
from enum import Enum
from typing import Optional

from pydantic import BaseModel


class Status(str, Enum):
    success = "success"
    error = "error"


class MsgResponse(BaseModel):
    msg: str
    status: Optional[Status] = None
    timestamp: datetime.datetime = datetime.datetime.now()



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, Union

from cryptography.fernet import Fernet

from ambient_client_common.utils import hash_string, logger


class TokenRepository(ABC):
    @abstractmethod
    def save_access_token(self, token: str) -> None:
        """Save the access token to the repository.

        Args:
            token (str): JWT access token
        """

    @abstractmethod
    def get_access_token(self) -> str:
        """Get the access token from the repository.

        Returns:
            str: JWT access token
        """

    @abstractmethod
    def save_refresh_token(self, token: str) -> None:
        """Save the refresh token to the repository.

        Args:
            token (str): JWT refresh token
        """

    @abstractmethod
    def get_refresh_token(self) -> str:
        """Get the refresh token from the repository.

        Returns:
            str: JWT refresh token
        """


class EncryptedTokenRepository(TokenRepository):
    def __init__(self, key: Optional[str] = None):
        self.key = key or Fernet.generate_key()
        self.cipher_suite = Fernet(self.key)
        self.home_path = Path.home() / ".ambient"
        self.access_token_path = self.home_path / ".access_token.enc"
        self.refresh_token_path = self.home_path / ".refresh_token.enc"
        self.fernet_key_path = self.home_path / f".{hash_string('fernet_key')}"

        self._init_files()

    def _init_files(self):
        self.home_path.mkdir(parents=True, exist_ok=True, mode=0o700)

        if not self.access_token_path.exists():
            logger.debug("creating access token file ...")
            self._write_to_file(self.access_token_path, b"")

        if not self.refresh_token_path.exists():
            logger.debug("creating refresh token file ...")
            self._write_to_file(self.refresh_token_path, b"")

        if not self.fernet_key_path.exists():
            logger.debug("creating fernet key file ...")
            self._write_to_file(self.fernet_key_path, self.key)
        else:
            logger.debug("reading fernet key file ...")
            self.key = self._read_from_file(self.fernet_key_path)
            self.cipher_suite = Fernet(self.key)

    def _write_to_file(self, path: Path, data: bytes):
        with path.open("wb") as file:
            path.chmod(0o600)
            file.write(data)

    def _read_from_file(self, path: Path) -> bytes:
        if not path.exists():
            logger.warning(f"File not found: {path}")
            return b""
        with path.open("rb") as file:
            return file.read()

    def save_access_token(self, token: str):
        encrypted_token = self.cipher_suite.encrypt(token.encode())
        self._write_to_file(self.access_token_path, encrypted_token)

    def get_access_token(self) -> Union[str, None]:
        encrypted_token = self._read_from_file(self.access_token_path)
        if not encrypted_token:
            return None
        return self.cipher_suite.decrypt(encrypted_token).decode()

    def save_refresh_token(self, token: str):
        encrypted_token = self.cipher_suite.encrypt(token.encode())
        self._write_to_file(self.refresh_token_path, encrypted_token)

    def get_refresh_token(self) -> str:
        logger.debug("reading refresh token ...")
        encrypted_token = self._read_from_file(self.refresh_token_path)
        if not encrypted_token:
            return ""
        return self.cipher_suite.decrypt(encrypted_token).decode()



import pathlib
from typing import Optional

from pydantic import BaseModel


class ServiceConfigBase(BaseModel):
    environment: Optional[dict] = None


def get_exec_start() -> str:
    exec_path = pathlib.Path.home() / ".ambient/.venv/bin/ambient_edge_server"
    return str(exec_path)


class ServiceConfigLinux(ServiceConfigBase):
    description: str = "Ambient Edge Server"
    user: str
    group: str
    exec_start: str = get_exec_start()



import docker

from ambient_client_common.models.docker_models import DockerRoleEnum
from ambient_client_common.repositories.docker_repo import DockerRepo


class DockerService:
    def __init__(self, client: docker.DockerClient, docker_repo: DockerRepo):
        self.client = client
        self.docker_repo = docker_repo

    def get_join_token(self, role: DockerRoleEnum) -> str:
        swarm_info = self.docker_repo.get_swarm_info()
        if role == DockerRoleEnum.Manager:
            return swarm_info.JoinTokens.Manager
        elif role == DockerRoleEnum.Worker:
            return swarm_info.JoinTokens.Worker
        else:
            raise ValueError("Role must be either MANAGER or WORKER")



import base64

from fastapi import APIRouter, Depends

from ambient_client_common.utils import logger
from ambient_edge_server.models.api import MsgResponse
from ambient_edge_server.services.port_service import PortService
from ambient_edge_server.services.service_manager import svc_manager

logger.debug("ports.py - finished imports")

router = APIRouter(prefix="/ports", tags=["ports"])


@router.get("/")
async def get_ports(port_svc: PortService = Depends(svc_manager.get_port_service)):
    return await port_svc.get_ports()


@router.get("/{port_id}")
async def get_port(
    port_id: int, port_svc: PortService = Depends(svc_manager.get_port_service)
):
    return await port_svc.get_port(port_id)


@router.get("/port-forwards")
async def get_port_forwards(
    port_svc: PortService = Depends(svc_manager.get_port_service),
):
    return await port_svc.get_port_forwards()


@router.post("/port-forwards/{port_id}")
async def forward_port(
    port_id: int,
    ws_url: str,
    port_svc: PortService = Depends(svc_manager.get_port_service),
):
    decoded_ws_url = base64.b64decode(ws_url.encode()).decode()
    logger.debug(f"decoded_ws_url: {decoded_ws_url}")
    await port_svc.forward_port(port_id, decoded_ws_url)
    return MsgResponse(msg="Port forwarded", status="success")



import asyncio
import getpass
import grp
import os
import pwd
import subprocess
from abc import ABC, abstractmethod
from typing import Optional, Set, Union

from result import Err, Ok, Result

from ambient_client_common.utils import logger
from ambient_edge_server import config
from ambient_edge_server.models.system_service import (
    ServiceConfigBase,
    ServiceConfigLinux,
)
from ambient_edge_server.repos.system_daemon_repo import SystemDaemonRepo


class SystemDaemonService(ABC):
    @abstractmethod
    async def install(self) -> Union[None, Err[str]]:
        """Initialize the service."""

    @abstractmethod
    async def save_config(self) -> Result[str, str]:
        """Save the service to the system."""

    @abstractmethod
    async def start(self) -> Result[str, str]:
        """Start the service."""

    @abstractmethod
    async def stop(self) -> Result[str, str]:
        """Stop the service."""

    @abstractmethod
    async def restart(self) -> Result[str, str]:
        """Restart the service."""

    @property
    @abstractmethod
    def status(self) -> Result[str, str]:
        """Get the status of the service.

        Returns:
            Result[str, str]: The status of the service.
                Ok if the service is running.
                Err if the service is not running.
        """

    @property
    @abstractmethod
    def service_config(self) -> ServiceConfigBase:
        """Return the service configuration."""


class LinuxDaemonService(SystemDaemonService):
    def __init__(self, system_daemon_repo: SystemDaemonRepo):
        self.system_daemon_repo = system_daemon_repo
        self._service_config: ServiceConfigLinux = self.build_service_config()
        self.tasks: Set[asyncio.Task] = set()

    async def install(self, env_vars: Optional[dict] = None) -> Optional[Err[str]]:
        """
        This should only be called once when the service is first installed
        on the system. It will render the service configuration and save it to
        the system. It exits upon completion.
        """

        logger.info("starting the installation process ...")
        if env_vars:
            logger.info("Env vars requested, adding to _service_config ...")
            _temp_service_config = self.service_config
            _temp_service_config.environment = env_vars
            self.service_config = _temp_service_config
            logger.debug(
                "LinuxDaemonService.install() - _temp_service_config: {}",
                _temp_service_config,
            )
        save_config_result = await self.save_config()
        if save_config_result.is_err():
            logger.error(
                "Failed to save the configuration: {}", save_config_result.err()
            )
            return Err(save_config_result.err())
        else:
            logger.info("Configuration saved successfully.")

        start_result = await self.start()
        if start_result.is_err():
            logger.error("Failed to start the service: {}", start_result.err())
            return start_result
        else:
            logger.info("Service started successfully.")

        logger.info("Installation complete. Exiting ...")
        return await self.restart()

    @property
    def service_config(self) -> ServiceConfigLinux:
        return self._service_config

    @service_config.setter
    def service_config(self, value: ServiceConfigLinux):
        self._service_config = value

    async def save_config(self) -> Result[str, str]:
        logger.info("Saving configuration ...")
        logger.debug("service config: {}", self.service_config)

        result: Result = await self.system_daemon_repo.render_template_and_save(
            self.service_config
        )
        if result.is_err():
            logger.error("Failed to save the configuration: {}", result.err())
            return result
        elif result.is_ok():
            logger.info("Configuration saved successfully.")
            return result

    async def start(self) -> Result[str, str]:
        logger.info("Starting the service ...")

        refresh_result = await self._refresh_services()
        if refresh_result.is_err():
            return refresh_result

        return await self._start_service()

    async def stop(self) -> Result[str, str]:
        logger.info("Stopping the service ...")
        try:
            stop_task = asyncio.create_task(
                self._run_systemctl_cmd("stop", delay=config.settings.default_delay)
            )
            logger.debug("LinuxDaemonService.stop() - stop_task: {}", stop_task)
            self.tasks.add(stop_task)
            stop_task.add_done_callback(lambda _: self.tasks.remove(stop_task))
            ok_msg = f"Service will stop in {config.settings.default_delay} seconds ..."
            logger.info(ok_msg)
            return Ok(ok_msg)
        except Exception as e:
            err_msg = f"Failed to stop the service: {e}"
            logger.error(err_msg)
            return Err(err_msg)

    async def restart(self) -> Result[str, str]:
        logger.info("Restarting the service ...")
        try:
            restart_task = asyncio.create_task(
                self._run_systemctl_cmd("restart", delay=config.settings.default_delay)
            )
            logger.debug(
                "LinuxDaemonService.restart() - restart_task: {}", restart_task
            )
            self.tasks.add(restart_task)
            restart_task.add_done_callback(lambda _: self.tasks.remove(restart_task))
            ok_msg = (
                f"Service will restart in {config.settings.default_delay} seconds ..."
            )
            logger.info(ok_msg)
            return Ok(ok_msg)
        except Exception as e:
            err_msg = f"Failed to restart the service: {e}"
            logger.error(err_msg)
            return Err(err_msg)

    @property
    def status(self) -> Result[str, str]:
        cmd = ["systemctl", "status", "ambient_edge_server.service"]

        logger.debug("command: {}", cmd)

        try:
            output = subprocess.run(cmd, capture_output=True, text=True)
            logger.debug("output: {}", output)
            output.check_returncode()
            logger.info("service is running.")
            return Ok("service is running.")
        except subprocess.CalledProcessError as e:
            logger.error("service is not running: {}", e)
            return Err("service is not running.")

    async def _refresh_services(self) -> Result[str, str]:
        refresh_cmd = ["systemctl", "daemon-reload"]

        logger.debug("refreshing services ...")
        logger.debug("cmd: {}", refresh_cmd)

        try:
            output = subprocess.run(refresh_cmd, capture_output=True, text=True)
            logger.debug("output: {}", output)
            output.check_returncode()
            logger.info("services refreshed successfully.")
            return Ok("services refreshed successfully.")
        except subprocess.CalledProcessError as e:
            logger.error("Failed to refresh services: {}", e)
            return Err("Failed to refresh services.")

    async def _start_service(self) -> Result[str, str]:
        return await self._run_systemctl_cmd("start")

    async def _run_systemctl_cmd(
        self, action: str, delay: Optional[int] = None
    ) -> Result[str, str]:

        cmd = ["systemctl", action, "ambient_edge_server.service"]

        logger.debug("command: {}", cmd)

        if delay:
            logger.info("Waiting {} seconds before running command ...", delay)
            await asyncio.sleep(delay)

        try:
            logger.info("running action: {} ...", action)
            output = subprocess.run(cmd, capture_output=True, text=True)
            logger.debug("output: {}", output)
            output.check_returncode()
            logger.info("action {} completed successfully.", action)
            return Ok(f"action {action} completed successfully.")
        except subprocess.CalledProcessError as e:
            logger.error("Failed to complete action: {}", e)
            return Err("Failed to complete action.")

    def build_service_config(
        self, env_vars: Optional[dict] = None
    ) -> ServiceConfigLinux:
        logger.debug("building the service configuration ...")
        user = self.get_user()
        group = self.get_group()

        if env_vars:
            logger.debug(
                "LinuxDaemonService.build_service_config - environment variables: {}",
                env_vars,
            )
        return ServiceConfigLinux(
            user=user,
            group=group,
            environment=env_vars,
        )

    def get_user(self) -> str:
        logger.debug("getting the current user ...")
        username = getpass.getuser()
        logger.info("current user: {}", username)
        return username

    def get_group(self) -> str:
        logger.debug("getting the current group ...")

        uid = os.getuid()
        user_info = pwd.getpwuid(uid)
        primary_gid = user_info.pw_gid

        logger.debug(
            "-> uid: {}\n-> primary_gid: {}\n-> user_info: {}",
            uid,
            primary_gid,
            user_info,
        )

        group_info = grp.getgrgid(primary_gid)
        group_name = group_info.gr_name

        logger.info("current group: {}", group_name)
        logger.debug("-> group_info: {}", group_info)

        return group_name



import ipaddress
import json
from abc import ABC, abstractmethod
from typing import List, Literal

import psutil
from ambient_backend_api_client.models import InterfaceTypeEnum, NetworkInterface
from result import Ok

from ambient_client_common.utils import logger


class InterfaceService(ABC):
    @abstractmethod
    async def get_network_interfaces(self) -> List[NetworkInterface]:
        """Get network interfaces on the device."""


class LinuxInterfaceService(InterfaceService):
    def __init__(self) -> None:
        self._running = False
        self._error = Ok("initialized")

    def determine_if_ipv4_or_ipv6(
        self, address: str
    ) -> Literal["ipv4", "ipv6", "unknown"]:
        try:
            ip_obj = ipaddress.ip_address(address)
        except ValueError as e:
            logger.error("Error parsing IP address: {}", e)
            return "unknown"
        if ip_obj.version == 4:
            return "ipv4"
        elif ip_obj.version == 6:
            return "ipv6"

    async def get_network_interfaces(self) -> List[NetworkInterface]:
        logger.info("Starting to fetch network interface details.")
        interfaces = psutil.net_if_addrs()
        logger.debug("Interfaces: {}", json.dumps(interfaces, indent=4))
        network_interfaces: List[NetworkInterface] = []

        for interface_name, addresses in interfaces.items():
            logger.debug("Processing {} interface ...", interface_name)
            logger.debug("Addresses: {}", addresses)

            for addr in addresses:
                existing_interface = next(
                    (
                        interface_
                        for interface_ in network_interfaces
                        if interface_.name == interface_name
                    ),
                    None,
                )
                if existing_interface:
                    logger.debug(
                        "Interface {} already exists in network interfaces.",
                        interface_name,
                    )
                    existing_addr = (
                        existing_interface.ipv4_address
                        or existing_interface.ipv6_address
                    )
                    if self.determine_if_ipv4_or_ipv6(existing_addr) == "ipv4":
                        if self.determine_if_ipv4_or_ipv6(addr.address) == "ipv6":
                            existing_interface.ipv6_address = (
                                addr.address.split("%")[0]
                                if "%" in addr.address
                                else addr.address
                            )
                    elif self.determine_if_ipv4_or_ipv6(existing_addr) == "ipv6":
                        if self.determine_if_ipv4_or_ipv6(addr.address) == "ipv4":
                            existing_interface.ipv4_address = addr.address
                    continue
                details = {
                    "name": f"{interface_name}",
                    # Default to OTHER, adjust logic as needed
                    "type": InterfaceTypeEnum.UNKNOWN,
                }
                logger.debug("Address: {}", addr)
                ip_version = self.determine_if_ipv4_or_ipv6(addr.address)
                if ip_version == "ipv4":
                    details["ipv4_address"] = addr.address
                    details["netmask"] = addr.netmask
                elif ip_version == "ipv6":
                    details["ipv6_address"] = addr.address
                else:
                    logger.error("Unknown IP version for address: {}", addr.address)
                    continue
                details["broadcast"] = addr.broadcast

                # Interface type heuristic
                wifi_heuristics = ["wi-fi", "wifi", "wlan"]
                ethernet_heuristics = [
                    "eth",
                    "enp",
                    "ens",
                    "eno",
                    "enx",
                    "docker",
                    "lo",
                    "br",
                ]

                if any(
                    heuristic in interface_name.lower() for heuristic in wifi_heuristics
                ):
                    details["type"] = InterfaceTypeEnum.WIFI
                elif any(
                    heuristic in interface_name.lower()
                    for heuristic in ethernet_heuristics
                ):
                    details["type"] = InterfaceTypeEnum.ETHERNET

                network_interface = NetworkInterface(**details)
                logger.debug(
                    "Network Interface: {}", network_interface.model_dump_json(indent=4)
                )
                network_interfaces.append(network_interface)
            logger.info(f"Added {interface_name} to network interfaces.")

        logger.info("Completed fetching all network interfaces.")
        return network_interfaces



from ambient_backend_api_client import Configuration
from ambient_event_bus_client import Client, ClientOptions
from docker import DockerClient

from ambient_client_common.repositories import docker_repo, node_repo
from ambient_client_common.utils import logger
from ambient_edge_server.config import settings
from ambient_edge_server.repos import plugin_repo, system_daemon_repo, token_repo
from ambient_edge_server.services import (
    authorization_service,
    crud_service,
    docker_service,
    event_service,
    health_service,
    interface_service,
    port_service,
    registry_service,
    system_daemon_service,
)


class ServiceManager:
    def __init__(self):
        if settings.platform == "linux":
            self._system_daemon_repo = system_daemon_repo.LinuxSystemDaemonRepo()
            self._system_daemon_service = system_daemon_service.LinuxDaemonService(
                self._system_daemon_repo
            )
            self._interface_service = interface_service.LinuxInterfaceService()
        else:
            logger.error("Platform not supported")
            raise NotImplementedError("Platform not supported")
        self._port_service = port_service.PortService()
        self._token_repo = token_repo.EncryptedTokenRepository()
        self._node_repo = node_repo.FileNodeRepo()
        self._authorization_service = authorization_service.AuthorizationService(
            token_repo=self._token_repo, node_repo=self._node_repo
        )
        self._docker_client = DockerClient()
        self._docker_repo = docker_repo.DockerRepo(client=self._docker_client)
        self._plugin_repo = plugin_repo.PluginRepo()

        self._docker_service = docker_service.DockerService(
            client=self._docker_client, docker_repo=self._docker_repo
        )

        self._event_client = Client(
            ClientOptions(
                event_api_url=settings.event_bus_api,
                connection_service_url=settings.connection_service_url,
                api_token="dummy_token",
                log_level=settings.ambient_log_level,
            )
        )
        self._event_service = event_service.AmbientBusEventService(
            client=self._event_client,
            node_repo=self._node_repo,
            plugin_repo=self._plugin_repo,
            docker_repo=self._docker_repo,
            auth_svc=self._authorization_service,
        )

        self._registry_svc_factory = registry_service.RegistryServiceFactory(
            docker_client=self._docker_client,
            auth_svc=self._authorization_service,
        )

        # Keep Health Service at the bottom since
        # it should be able to access all other services
        if settings.platform == "linux":
            self._health_service = health_service.LinuxHealthService(
                node_repo=self._node_repo,
                event_service=self._event_service,
                auth_service=self._authorization_service,
                interface_service=self._interface_service,
                docker_repo=self._docker_repo,
                interval_minutes=5,
            )

        self._crud_service = crud_service.CRUDService(node_repo=self._node_repo)

    async def init(self):
        logger.info("Async Initializing services ...")
        await self._port_service.init()
        logger.debug("Port service initialized")

        logger.debug("Fetching token ...")
        token = await self._authorization_service.get_token()
        if not token:
            logger.error("Failed to get token")
            return
        logger.debug("got token [ {} chars ]", len(token))
        self._event_client.token = token
        logger.debug("Event client token set")
        await self._event_client.init_client()
        logger.debug("Event client initialized")

        api_config = Configuration(
            host=settings.backend_api_url,
            access_token=token,
        )
        logger.info(
            "API Config initialized. Host: {} -- Access Token: {}",
            api_config.host,
            api_config.access_token,
        )

        await self._event_service.start(api_config=api_config)
        logger.debug("Event service started")

        await self._health_service.start(api_config=api_config)
        logger.debug("Health service started")
        await self._crud_service.init(api_config=api_config)
        logger.debug("CRUD service initialized")
        await self._registry_svc_factory.init()
        logger.debug("Registry service factory initialized")

    def get_port_service(self):
        return self._port_service

    def get_authorization_service(self):
        return self._authorization_service

    def get_event_service(self):
        return self._event_service

    def get_system_daemon_service(self):
        return self._system_daemon_service

    def get_interface_service(self):
        return self._interface_service

    def get_crud_service(self):
        return self._crud_service

    def get_health_service(self):
        return self._health_service

    def get_docker_service(self):
        return self._docker_service


svc_manager = ServiceManager()



import datetime
import pathlib

from fastapi import FastAPI

from ambient_client_common.utils import logger
from ambient_edge_server import config
from ambient_edge_server.routers import auth, crud, daemon, health, ping, plugins, ports
from ambient_edge_server.services.service_manager import svc_manager

app = FastAPI()
logger.debug("Initialized FastAPI app")


async def startup():
    logger.info("Starting up ...")
    # initialize services
    try:
        await svc_manager.init()
        logger.info("Services initialized.")

    except Exception as e:
        logger.warning("Failed to initialize services: {}", e)

    # ensure asset availabilitry
    try:
        package_location = pathlib.Path(__file__).parent
        logger.debug("app.startup() - __file__: {}", __file__)
        logger.debug("app.startup() - package location: {}", package_location)
        jinja_template_location = (
            package_location / config.settings.service_template_location
        )
        if not jinja_template_location.exists():
            logger.warning("Jinja template file not found: {}", jinja_template_location)
        word_art_path = package_location / config.settings.word_art_path
        if not word_art_path.exists():
            logger.warning("Word art file not found: {}", word_art_path)
            return
        logger.debug("Word Art path: {}", word_art_path)
        logger.info("\n{}\n", open(word_art_path.absolute()).read())
        logger.info(
            "Started at: {}\t\t\tVersion:{}",
            datetime.datetime.now(),
            config.settings.version,
        )
    except Exception as e:
        logger.error("Failed to load assets: {}", e)


app.add_event_handler("startup", startup)

routers = [ping, ports, auth, daemon, crud, health, plugins]

for router in routers:
    app.include_router(router.router)



from abc import ABC, abstractmethod

from ambient_backend_api_client import EventLabel
from ambient_event_bus_client import Message

from ambient_edge_server.services.event_service import EventService


class BaseHandler(ABC):
    def __init__(self, event_service: EventService):
        self.event_svc = event_service

    @abstractmethod
    async def handle(self, msg: Message) -> None:
        """Handle a msg.

        Args:
            msg (Message): Message to handle
        """

    @property
    @abstractmethod
    def label(self) -> EventLabel:
        """Get the label of the handler.

        Returns:
            str: Handler label
        """

    async def subscribe(self):
        await self.event_svc.add_event_handler(self.label, self.handle)



from ambient_backend_api_client import NodeOutput as Node
from fastapi import APIRouter, Depends, HTTPException

from ambient_client_common.models.docker_models import DockerRoleEnum
from ambient_edge_server.services import crud_service, docker_service, service_manager

router = APIRouter(prefix="/data", tags=["data", "nodes"])


# GET node data
@router.get("/node", response_model=Node)
async def get_node_data(
    crud_service: crud_service.CRUDService = Depends(
        service_manager.svc_manager.get_crud_service
    ),
):
    node_data = await crud_service.get_node_data()
    if not node_data:
        raise HTTPException(status_code=404, detail="Node data not found")
    return node_data


# PUT node data
@router.put("/node", response_model=Node)
async def update_node_data(
    crud_service: crud_service.CRUDService = Depends(
        service_manager.svc_manager.get_crud_service
    ),
):
    node_data = await crud_service.update_node_data()
    return node_data


# DELETE node data
@router.delete("/node")
async def delete_node_data(
    crud_service: crud_service.CRUDService = Depends(
        service_manager.svc_manager.get_crud_service
    ),
):
    await crud_service.clear_node_data()
    return {"message": "Node data deleted"}


# GET Docker swarm join token
@router.get("/swarm/join-token")
async def get_swarm_join_token(
    role: DockerRoleEnum,
    docker_svc: docker_service.DockerService = Depends(
        service_manager.svc_manager.get_docker_service
    ),
):
    join_token = docker_svc.get_join_token(role=role)
    return {"join_token": join_token}



import pathlib
from typing import Any, Dict, List, Union

import yaml
from ambient_base_plugin import PluginDefinition
from pydantic import ValidationError

from ambient_client_common.utils import logger
from ambient_edge_server.config import settings


class PluginRepo:
    def load_plugins(
        self, extra_plugin_paths: List[pathlib.Path] = []
    ) -> List[PluginDefinition]:
        """Load plugins

        Args:
            extra_plugin_paths (List[pathlib.Path], optional):
                Extra plugin paths. Defaults to [].

        Returns:
            List[PluginDefinition]: List of plugins
        """
        logger.info("Loading plugins ...")
        builtin_plugins = self.get_builtin_plugins()
        logger.debug("Builtin plugins: {}", builtin_plugins)
        extra_plugins = []
        for path in extra_plugin_paths:
            extra_plugins.extend(self.get_plugins_from_path(path))
        logger.debug("Extra plugins: {}", extra_plugins)
        all_plugins = builtin_plugins + extra_plugins
        logger.info("Loaded {} plugins", len(all_plugins))
        return all_plugins

    def get_builtin_plugins(self) -> List[PluginDefinition]:
        """Get builtin plugins

        Returns:
            List[PluginDefinition]: List of builtin plugins
        """
        logger.info("Loading builtin plugins ...")
        logger.debug("Settings package location: {}", settings.package_location)
        builtin_plugins_path = settings.package_location / "builtin_plugins.yml"
        logger.debug("Builtin plugins path: {}", builtin_plugins_path)

        builtin_plugins = self.get_plugins_from_path(builtin_plugins_path)
        logger.info("Loaded {} builtin plugins", len(builtin_plugins))
        return builtin_plugins

    def load_yaml(self, path: pathlib.Path) -> Union[List[Dict[str, Any]], None]:
        """Load a yaml file

        Args:
            path (pathlib.Path): Path to yaml file

        Returns:
            Union[List[Dict[str, Any]], None]: List of dictionaries or None
        """

        try:
            with open(path, "r") as file:
                return yaml.safe_load(file)
        except Exception as e:
            logger.warning("Error loading yaml file: {}", e)
            return None

    def get_plugins_from_path(self, path: pathlib.Path) -> List[PluginDefinition]:
        """Get plugins from a path

        Args:
            path (pathlib.Path): Path to load plugins from

        Returns:
            List[PluginDefinition]: List of plugins
        """

        logger.info("Loading plugins from path: {}", path)
        if not path.exists():
            logger.warning("Path does not exist: {}", path)
            return []
        plugins = self.load_yaml(path)
        if plugins is None:
            logger.warning("No plugins loaded from path: {}")
            return []
        parsed_plugins = []
        for plugin in plugins:
            try:
                parsed_plugins.append(PluginDefinition.model_validate(plugin))
            except ValidationError as e:
                logger.error("Error parsing plugin: {}", e)
        logger.info("Loaded {} plugins from path: {}", len(parsed_plugins), path)
        return parsed_plugins



import platform
from pathlib import Path

from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    version: str = "1.2.3"
    ambient_log_level: str = "INFO"
    ambient_log_lines: int = 1000
    ambient_config: str = "~/.ambientctl/config.json"

    backend_api_url: str = "https://api.ambientlabs.io"
    event_bus_api: str = "https://events.ambientlabs.io"
    connection_service_url: str = "wss://sockets.ambientlabs.io"

    package_location: Path = Path(__file__).parent
    service_template_location: str = "./templates/ambient_edge_server.service.jinja2"
    platform: str = platform.system().lower()

    word_art_path: str = "./assets/word_art.txt"
    default_delay: int = 3

    log_file_size: str = "10 MB"


settings = Settings()



import os
import sys

import uvicorn


def run():
    print(f"sys.path: {sys.path}    os.getcwd(): {os.getcwd()}")
    try:
        print("attempting imports ...")
        from ambient_edge_server.app import app

        print(f"app: {app}")
        from ambient_client_common.utils import logger

        print(f"logger: {logger}")

        logger.debug("running server ...")
        logger.debug("running app: {} ...", app)
        uvicorn.run(app, host="0.0.0.0", port=7417)
    except Exception as e:
        print("error running server: ", e)


if __name__ == "__main__":
    run()



import hashlib


def hash_string(input_string: str) -> str:
    # Create a new sha256 hash object
    hash_object = hashlib.sha256()

    # Encode the input string and update the hash object
    hash_object.update(input_string.encode("utf-8"))

    # Get the hexadecimal representation of the hash
    hashed_string = hash_object.hexdigest()

    return hashed_string



import aiohttp
from fastapi import APIRouter

from ambient_client_common.utils import logger
from ambient_edge_server.config import settings
from ambient_edge_server.models.api import MsgResponse

logger.debug("ping.py - finished imports")

router = APIRouter()


@router.get("/ping", response_model=MsgResponse)
async def pong(endpoint: str) -> MsgResponse:
    if endpoint == "server":
        return MsgResponse(msg="pong", status="success")
    elif endpoint == "backend":
        async with aiohttp.ClientSession() as session:
            url = f"{settings.backend_api_url}/ping/"
            logger.debug("url: {}", url)
            async with session.get(url) as resp:
                resp.raise_for_status()
                return MsgResponse(msg="pong", status="success")



from typing import Optional

from ambient_backend_api_client import ApiClient, Configuration
from ambient_backend_api_client import NodeOutput as Node
from ambient_backend_api_client import NodesApi

from ambient_client_common.repositories.node_repo import NodeRepo
from ambient_client_common.utils import logger


class CRUDService:
    """A service for interacting with stateful data on the node"""

    def __init__(self, node_repo: NodeRepo) -> None:
        self.node_repo = node_repo
        self.api_config: Optional[Configuration] = None

    async def init(self, api_config: Configuration) -> None:
        """Initialize the service with API configuration

        Args:
            api_config (Configuration): API configuration
        """
        self.api_config = api_config

    async def get_node_data(self) -> Node:
        """Get the node data from the repository.

        Returns:
            Node: Node data
        """
        logger.info("Fetching Node data from repository ...")
        return self.node_repo.get_node_data()

    async def clear_node_data(self) -> None:
        """Clear the node data in the repository.

        Returns:
            None
        """
        logger.info("Clearing Node data from repository ...")
        self.node_repo.clear_node_data()
        logger.info("Node data cleared")

    async def update_node_data(self) -> Node:
        """Fetch Node data from API and save to repo

        Returns:
            Node: Node data
        """

        logger.info("Updating Node data ...")
        # get node ID from repo
        node_id = self.node_repo.get_node_id()
        logger.debug("CRUDService.update_node_data() - Node ID: {}", node_id)

        # make API call
        if not self.api_config:
            logger.error("API Configuration not found")
            raise Exception("Has the CRUDService been initiated?")
        node: Optional[Node] = None
        async with ApiClient(self.api_config) as api_client:
            nodes_api = NodesApi(api_client)
            logger.debug(
                "CRUDService.update_node_data() - Fetching Node data from API ..."
            )
            node = await nodes_api.get_node_nodes_node_id_get(node_id=node_id)
            logger.debug("CRUDService.update_node_data() - Node data fetched: {}", node)
        logger.info("Node data fetched")

        # save data using node repo
        logger.info("Saving Node data to repository ...")
        self.node_repo.save_node_data(node)
        logger.info("Node data saved")

        # return node data
        return node



from typing import Optional

import fastapi

from ambient_client_common.utils import logger
from ambient_edge_server.services.event_service import EventService
from ambient_edge_server.services.service_manager import svc_manager

logger.debug("plugins.py - finished imports")

router = fastapi.APIRouter(tags=["plugins"])
logger.debug("plugins.py - created router")


@router.post("/plugins/{plugin_name}/{custom_endpoint}")
async def handle_custom_plugin_endpoint_post(
    plugin_name: str,
    custom_endpoint: str,
    request: fastapi.Request,
    body: Optional[dict] = None,
    event_service: EventService = fastapi.Depends(svc_manager.get_event_service),
):
    return await event_service.handle_plugin_api_request("post", request.url.path, body)


@router.get("/plugins/{plugin_name}/{custom_endpoint}")
async def handle_custom_plugin_endpoint_get(
    plugin_name: str,
    custom_endpoint: str,
    request: fastapi.Request,
    event_service: EventService = fastapi.Depends(svc_manager.get_event_service),
):
    return await event_service.handle_plugin_api_request("get", request.url.path)


@router.put("/plugins/{plugin_name}/{custom_endpoint}")
async def handle_custom_plugin_endpoint_put(
    plugin_name: str,
    custom_endpoint: str,
    request: fastapi.Request,
    body: Optional[dict] = None,
    event_service: EventService = fastapi.Depends(svc_manager.get_event_service),
):
    return await event_service.handle_plugin_api_request("put", request.url.path, body)


@router.delete("/plugins/{plugin_name}/{custom_endpoint}")
async def handle_custom_plugin_endpoint_delete(
    plugin_name: str,
    custom_endpoint: str,
    request: fastapi.Request,
    event_service: EventService = fastapi.Depends(svc_manager.get_event_service),
):
    return await event_service.handle_plugin_api_request("delete", request.url.path)


logger.debug("plugins.py - created routes")



from fastapi import APIRouter, Depends

from ambient_client_common.utils import logger
from ambient_edge_server.models.api import MsgResponse
from ambient_edge_server.services.health_service import HealthService
from ambient_edge_server.services.service_manager import svc_manager

logger.debug("health.py - finished imports")

router = APIRouter(prefix="/health", tags=["health"])


@router.get("/check-in", response_model=MsgResponse)
async def check_in(
    health_svc: HealthService = Depends(svc_manager.get_health_service),
):
    """Check in with the backend API

    Args:
        health_svc (HealthService, optional): Defaults to
            Depends(svc_manager.get_health_service).

    Returns:
        dict: Response
    """

    await health_svc.trigger_check_in()
    return MsgResponse(
        msg="Triggered Check-In",
        status="success",
    )



from typing import Optional

from fastapi import APIRouter, Depends, HTTPException

from ambient_edge_server.models.api import MsgResponse
from ambient_edge_server.services.service_manager import svc_manager
from ambient_edge_server.services.system_daemon_service import SystemDaemonService

router = APIRouter(prefix="/daemon", tags=["daemon"])


@router.post("/install", response_model=None)
async def install_service(
    env_vars: Optional[dict] = None,
    daemon_svc: SystemDaemonService = Depends(svc_manager.get_system_daemon_service),
):
    result = await daemon_svc.install(env_vars)
    if result.is_err():
        raise HTTPException(status_code=500, detail=result.unwrap_err())

    # we shouldn't event get here because this endpoint
    # will stop the server (to install the service)
    return None


@router.post("/start", response_model=MsgResponse)
async def start_service(
    daemon_svc: SystemDaemonService = Depends(svc_manager.get_system_daemon_service),
):
    result = await daemon_svc.start()
    if result.is_err():
        raise HTTPException(status_code=500, detail=result.err())

    return MsgResponse(msg="Service started.")


@router.post("/stop", response_model=MsgResponse)
async def stop_service(
    daemon_svc: SystemDaemonService = Depends(svc_manager.get_system_daemon_service),
):
    result = await daemon_svc.stop()
    if result.is_err():
        return MsgResponse(msg=result.unwrap_err(), status="error")
    return MsgResponse(msg=result.unwrap(), status="success")


@router.post("/restart", response_model=MsgResponse)
async def restart_service(
    daemon_svc: SystemDaemonService = Depends(svc_manager.get_system_daemon_service),
):
    result = await daemon_svc.restart()
    if result.is_err():
        return MsgResponse(msg=result.unwrap_err(), status="error")
    return MsgResponse(msg=result.unwrap(), status="success")


@router.get("/status", response_model=MsgResponse)
async def service_status(
    daemon_svc: SystemDaemonService = Depends(svc_manager.get_system_daemon_service),
):
    status = daemon_svc.status
    if status.is_err():
        raise HTTPException(status_code=500, detail=status.err())

    return MsgResponse(msg=status.unwrap())



import json
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional

import docker
import websockets
from ambient_backend_api_client import (
    ApiClient,
    Configuration,
    ContainerRegistry,
    ContainerRegistryAuth,
    ContainerRegistryType,
    RegistriesApi,
)
from cryptography.fernet import Fernet
from result import Err, Ok, Result

from ambient_client_common.utils import logger
from ambient_edge_server import config
from ambient_edge_server.services.authorization_service import AuthorizationService


class RegistryService(ABC):
    @abstractmethod
    async def authorize(self, ws_session: str) -> Result[None, str]:
        """
        Authorize the registry.

        Args:
            ws_session (str): Websocket session
            registry_type (ContainerRegistryType): Registry type

        Returns:
            Result[None, str]: Ok if authorized, Err otherwise
        """
        raise NotImplementedError("Method not implemented")


class RegistryServiceBase:
    """A base class for registry service to perform repeated tasks"""

    def __init__(self, auth_svc: AuthorizationService) -> None:
        super().__init__()
        self.auth_svc = auth_svc
        logger.debug("RegistryServiceBase initialized.")
        self.api_config: Optional[Configuration] = None

    async def init(self) -> None:
        self.api_config = Configuration(
            host=config.settings.backend_api_url,
            access_token=await self.auth_svc.get_token(),
        )

    async def _get_registry(self, registry_id: int) -> ContainerRegistry:
        """Get Container Registry from API

        Args:
            registry_id (int): Registry ID

        Returns:
            ContainerRegistry: ContainerRegistry object
        """
        async with ApiClient(self.api_config) as api_client:
            logger.debug("RegistryServiceBase._get_registry() - within async context")
            registries_api = RegistriesApi(api_client)
            logger.info(
                "Making request to backend API with registry ID: {}", registry_id
            )
            return await registries_api.get_registry_registries_registry_id_get_0(
                registry_id=registry_id
            )

    async def _get_registry_creds(self, creds_id: int) -> ContainerRegistryAuth:
        """Get Container Registry Auth from API

        Args:
            creds_id (int): Credentials ID

        Returns:
            ContainerRegistryAuth: ContainerRegistryAuth object
        """
        async with ApiClient(self.api_config) as api_client:
            logger.debug(
                "RegistryServiceBase._get_registry_creds() - within async context"
            )
            registries_api = RegistriesApi(api_client)
            logger.info("Making request to backend API with creds ID: {}", creds_id)
            return await registries_api.get_registry_auth_cred_registries_creds_creds_id_get(  # noqa
                creds_id=creds_id
            )

    async def _decrypt_password(self, creds: ContainerRegistryAuth) -> str:
        """Decrypt the password

        Args:
            creds (ContainerRegistryAuth): ContainerRegistryAuth object

        Returns:
            str: Decrypted password
        """
        logger.info("Decrypting password ...")
        encrypted_password = creds.password.encode("utf-8")
        logger.debug(
            "RegistryServiceBase._decrypt_password() - encrypted password encoded [ {} chars ]",  # noqa
            len(encrypted_password),
        )
        cipher_suite: Fernet = self.auth_svc.get_cipher_suite()
        decrypted_password = cipher_suite.decrypt(encrypted_password).decode("utf-8")
        logger.debug(
            "RegistryServiceBase._decrypt_password() - decrypted password [ {} chars ]",
            len(decrypted_password),
        )
        logger.info("Password decrypted successfully")
        return decrypted_password


class DockerRegistryService(RegistryService, RegistryServiceBase):
    def __init__(self, docker_client: docker.DockerClient, **kwargs) -> None:
        super().__init__(**kwargs)
        self.docker_client = docker_client
        logger.debug("DockerRegistryService initialized.")

    async def authorize(self, ws_session: str) -> Result[None, str]:
        """
        Authorize the registry.

        Args:
            ws_session (str): Websocket session
            registry_type (ContainerRegistryType): Registry type

        Returns:
            Result[None, str]: Ok if authorized, Err otherwise
        """
        logger.info("Authorizing to registry ...")
        logger.debug("DockerRegistryService.authorize() - ws_session: {}", ws_session)
        try:
            # connect to ws session
            async with websockets.connect(ws_session) as ws:
                logger.info("connected to web socket session")
                # send token
                token = await self.auth_svc.get_token()
                await ws.send(token)
                logger.info("authorized web socket session")

                # receive registry auth data
                auth_msg = await ws.recv()
                logger.debug(
                    "DockerRegistryService.authorize() - auth_msg: {}", auth_msg
                )
                auth_data: dict = json.loads(auth_msg)
                logger.debug(
                    "DockerRegistryService.authorize() - auth_data contains [ {} ] keys",  # noqa
                    len(auth_data.keys()),
                )
                creds_id = auth_data.get("creds_id", None)
                logger.info("Authorizing using credentials: {}", creds_id)
                if creds_id is None:
                    logger.error("creds_id not found in auth data")
                    return Err("creds_id not found in auth data")
                creds = await self._get_registry_creds(creds_id)
                logger.info("Got credentials for registry: {}", creds.registry_id)
                logger.debug(
                    "DockerRegistryService.authorize() - creds: {}",
                    creds.model_dump_json(indent=4),
                )
                registry = await self._get_registry(registry_id=creds.registry_id)
                logger.info("Got registry: {}", registry.id)
                logger.debug(
                    "DockerRegistryService.authorize() - registry: {}",
                    registry.model_dump_json(indent=4),
                )

                # send "ack"
                await ws.send("ack")
                logger.info("Acknowledged the auth request")

                # authorize to registry using the docker client
                url = registry.url
                username = creds.username
                logger.info(
                    "Authentication to registry @ [ {} ] using username: {} ...",
                    url,
                    username,
                )
                password = await self._decrypt_password(creds)
                logger.debug("DockerRegistryService.authorize() - password decrypted.")
                result = await self._docker_login(url, username, password)
                del password
                logger.debug("DockerRegistryService.authorize() - password deleted.")
                if result.is_err():
                    return result
                logger.info("Authenticated to registry [ {} ] successfully", url)

                # send "completed" message
                await ws.send("completed")
                logger.info("backend notified about successful authorization")

                # successful authorization
                logger.debug("DockerRegistryService.authorize() - returning Ok object")
                return Ok(None)
        except Exception as e:
            error_msg = f"Failed to authorize to registry: {e}"
            logger.error(error_msg)
            return Err(error_msg)

    async def _docker_login(
        self, url: str, username: str, password: str
    ) -> Result[Dict[Any, Any], str]:
        """Login to the docker registry.

        Args:
            url (str): URL of the registry
            username (str): username
            password (str): password (already decrypted)

        Returns:
            Result[Dict[Any, Any], str]: The authenticatoin response
        """
        try:
            logger.debug(
                "DockerRegistryService._docker_login() - url: {}, username: {}",
                url,
                username,
            )
            login_resp = self.docker_client.login(
                username=username, password=password, registry=url
            )
            logger.debug(
                "DockerRegistryService._docker_login() - login_resp: {}", login_resp
            )
            return Ok(login_resp)
        except docker.errors.APIError as e:
            error_msg = f"Failed to authenticate to registry (Docker API Error): {e}"
            logger.error(error_msg)
            return Err(error_msg)
        except Exception as e:
            error_msg = f"Failed to authenticate to registry: {e}"
            logger.error(error_msg)
            return Err(error_msg)


class RegistryServiceFactory:
    def __init__(
        self,
        docker_client: docker.DockerClient,
        auth_svc: AuthorizationService,
    ) -> None:
        base_kwargs = {
            "auth_svc": auth_svc,
        }
        self.docker_registry_service = DockerRegistryService(
            docker_client=docker_client, **base_kwargs
        )

    async def init(self):
        await self.docker_registry_service.init()

    def get_registry_service(
        self, registry_type: ContainerRegistryType
    ) -> RegistryService:
        if registry_type == ContainerRegistryType.DOCKER_HUB:
            return self.docker_registry_service
        else:
            raise NotImplementedError(f"Registry type {registry_type} not supported")



import asyncio
import socket
import subprocess
from typing import Any, Dict, Set

import websockets
from cryptography.fernet import Fernet

from ambient_client_common.utils import logger


class PortService:
    """
    Currently not used
    """

    def __init__(self) -> None:
        self._forwarded_ports = {}
        self._server_processes = {}
        self.tasks: Set[asyncio.Task] = set()

    async def init(self) -> None:
        pass

    async def cleanup(self) -> None:
        logger.info("Cleaning up PortService ...")
        # Terminate any active SSH processes
        process: subprocess.Popen
        for process in self._server_processes.values():
            process.terminate()
        self._server_processes.clear()
        logger.info("PortService cleanup complete")

    async def get_ports(self) -> Dict[int, str]:
        pass

    async def get_port(self, port: int) -> Any:
        pass

    async def get_port_forwards(self) -> Dict[int, str]:
        pass

    async def forward_port(self, port: int, websocket_url: str) -> None:
        logger.info(f"Forwarding port {port} to {websocket_url} ...")
        task = asyncio.create_task(self.forward_port_task(port, websocket_url))
        self.tasks.add(task)
        task.add_done_callback(lambda task: self.tasks.remove(task))

    async def forward_port_task(self, port: int, websocket_url: str) -> None:
        try:
            # Connect to the WebSocket server
            async with websockets.connect(websocket_url) as websocket:
                self._forwarded_ports[port] = websocket_url
                logger.debug(
                    "added port {} to forwarded_ports [ws url: {} ]",
                    port,
                    self.forwarded_ports.get(port),
                )

                # Receive the encryption key from the server
                key = await websocket.recv()
                fernet = Fernet(key.encode())
                logger.debug("created fernet key [ {} chars ]", len(key))

                # Forward traffic between WebSocket and local port (asynchronous)
                logger.info("forwarding traffic between WebSocket and local port ...")

                async def send_to_websocket(reader: asyncio.StreamReader) -> None:
                    while True:
                        data = await reader.read(1024)
                        logger.debug(
                            "received data from local port [ {} bytes ]", len(data)
                        )
                        if not data:
                            break
                        await websocket.send(fernet.encrypt(data))

                async def receive_from_websocket(writer: asyncio.StreamWriter) -> None:
                    async for message in websocket:
                        logger.debug(
                            "sending data to local port [ {} bytes]", len(message)
                        )
                        writer.write(fernet.decrypt(message))
                        await writer.drain()

                with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                    sock.bind(("localhost", port))  # Bind to the local port
                    sock.listen(1)
                    conn, addr = sock.accept()
                    with conn:
                        reader, writer = asyncio.StreamReader(), asyncio.StreamWriter(
                            conn, None, None, True
                        )
                        await asyncio.gather(
                            send_to_websocket(reader), receive_from_websocket(writer)
                        )

        except Exception as e:
            print(f"Error forwarding port {port}: {e}")

    @property
    def forwarded_ports(self) -> Dict[int, str]:
        """Currently forwarded ports

        Returns:
            Dict[int, str]: Dictionary of shape {port: websocket_url}
        """
        return self._forwarded_ports

    @forwarded_ports.setter
    def forwarded_ports(self, value: Dict[int, str]) -> None:
        self._forwarded_ports = value



import asyncio
import datetime
import json
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Set

import aiohttp
import apluggy as pluggy
from ambient_backend_api_client import ApiClient, Configuration, DockerSwarmInfo
from ambient_backend_api_client import NodeOutput as Node
from result import Err, Ok, Result

from ambient_client_common.repositories.docker_repo import DockerRepo
from ambient_client_common.repositories.node_repo import NodeRepo
from ambient_client_common.utils import logger
from ambient_edge_server.services.authorization_service import AuthorizationService
from ambient_edge_server.services.event_service import EventService
from ambient_edge_server.services.interface_service import InterfaceService


class HealthService(ABC):
    @abstractmethod
    async def start(self):
        """Start the service."""

    @abstractmethod
    async def stop(self):
        """Stop the service."""

    @abstractmethod
    async def get_health(self) -> str:
        """Get the health of the service."""

    @abstractmethod
    async def run_system_sweep(self) -> str:
        """Run a system sweep and update backend with results."""

    @property
    @abstractmethod
    def interval_min(self) -> int:
        """Get the interval in minutes for the system sweep."""


class LinuxHealthService(HealthService):
    def __init__(
        self,
        node_repo: NodeRepo,
        event_service: EventService,
        auth_service: AuthorizationService,
        interface_service: InterfaceService,
        docker_repo: DockerRepo,
        interval_minutes: int = 5,
    ):
        self.node_repo = node_repo
        self.event_service = event_service
        self.auth_service = auth_service
        self.interface_service = interface_service
        self.docker_repo = docker_repo
        self._interval_min_ = interval_minutes

        self.api_config: Optional[Configuration] = None
        self._running = False
        self._tasks: Set[asyncio.Task] = set()

    async def start(self, api_config: Configuration):
        self.api_config = api_config
        self._running = True
        sys_sweep_task = asyncio.create_task(self._system_sweep_task())
        sys_sweep_task.add_done_callback(self._handleDoneTask)
        self._tasks.add(sys_sweep_task)

    async def stop(self):
        self._running = False
        for task in self._tasks:
            task.cancel()
        self._tasks.clear()

    async def get_health(self) -> Result[str, str]:
        if not self._running:
            return Err("Service not running")
        if len(self._tasks) == 0:
            return Err("No health tasks running")
        return Ok("OK")

    async def run_system_sweep(self) -> Result[str, str]:
        """Run a system sweep

        Workflow:
            - Check health of the system
            - Check if event service is running
            - Check if node is authorized with backend

        Returns:
            Result[str, str]: Ok if system sweep completed successfully.
        """
        logger.info("Running system sweep ...")
        self_health_result = await self.get_health()
        logger.debug("self_health_result: {}", self_health_result)
        if self_health_result.is_err():
            logger.error(
                "Health check failed. Error: {}", self_health_result.unwrap_err()
            )
            return Err("Health check failed")

        if not self.event_service.is_running or self.event_service.error:
            logger.error(
                "Event service not running or in error state.\n\
  is_running: {}\n    error: {}",
                self.event_service.is_running,
                self.event_service.error,
            )
            return Err("Event service not running or in error state")

        if (await self.auth_service.verify_authorization_status()).is_err():
            logger.error("Node is not authorized with backend")
            return Err("Node is not authorized with backend")

        logger.info("Core system sweep completed successfully")

        logger.info("Running plugin system sweep ...")
        try:
            node = self.node_repo.get_node_data()
            await trigger_system_sweep_hook(
                plugin_manager=self.event_service.plugin_manager, node=node
            )
        except Exception as e:
            logger.error("Failed to run plugin system sweep: {}", e)

        return Ok("System sweep completed successfully")

    @property
    def interval_min(self) -> int:
        return self._interval_min_

    def _get_api_session(self) -> ApiClient:
        if not self.api_config:
            raise ValueError("API Configuration not set")
        return ApiClient(configuration=self.api_config)

    async def _system_sweep_task(self):
        while True:
            try:
                logger.debug("waiting for system sweep interval delay ... ")
                await asyncio.sleep(self._interval_min_ * 60)
                logger.debug("running system sweep ...")
                result = await self.run_system_sweep()
            except Exception as e:
                logger.error("error running system sweep: {}", str(e))
                result = Err(str(e))
            finally:
                logger.info("result of system sweep: {}", result)
                logger.debug("handling system sweep result ...")
                await self._handle_system_sweep_result(result)

    async def trigger_check_in(self):
        result = await self.run_system_sweep()
        await self._handle_system_sweep_result(result)

    async def _handle_system_sweep_result(self, result: Result[str, str]):
        """Handle the result of the system sweep

        This is the meat and potatoes of collecting the system information
        and updating the backend with the results.

        Args:
            result (Result[str, str]): Result of the system sweep.
        """
        logger.info("Handling system sweep result ...")
        last_seen = datetime.datetime.now().isoformat()
        logger.debug("last_seen: {}", last_seen)
        current_interfaces = await self.interface_service.get_network_interfaces()
        logger.debug("Found {} interfaces", len(current_interfaces))
        interfaces = [i.model_dump_json(indent=4) for i in current_interfaces]
        docker_info = self.docker_repo.get_docker_info()
        node_docker_info = None
        if docker_info.Swarm.Cluster:
            node_docker_info = DockerSwarmInfo(
                NodeID=docker_info.Swarm.NodeID,
                NodeAddr=docker_info.Swarm.NodeAddr,
                LocalNodeState=docker_info.Swarm.LocalNodeState,
                Error=docker_info.Swarm.Error,
            )
        if result.is_ok():
            logger.debug("updating backend with system sweep results ...")
            await self._patch_node(
                {
                    "last_seen": last_seen,
                    "status": "active",
                    "interfaces": interfaces,
                    "docker_swarm_info": (
                        node_docker_info.model_dump() if node_docker_info else None
                    ),
                }
            )
        else:
            await self._patch_node(
                {
                    "last_seen": last_seen,
                    "status": "error",
                    "error": result.unwrap_err(),
                    "interfaces": interfaces,
                }
            )

    async def _patch_node(self, data: Dict[str, Any]) -> Result[str, str]:
        async def __patch_node():
            logger.debug("in __patch_node ...")
            async with aiohttp.ClientSession() as session:
                async with session.patch(
                    f"{self.api_config.host}/nodes/{node_id}",
                    json=data,
                    headers={
                        "Authorization": f"Bearer {await self.auth_service.get_token()}"
                    },
                ) as response:
                    resp_text = None
                    try:
                        resp_text = await response.text()
                        response.raise_for_status()
                        updated_node_d = json.loads(resp_text)
                        updated_node = Node.model_validate(updated_node_d)
                        logger.debug(
                            "Node patched: {}", updated_node.model_dump_json(indent=4)
                        )
                        self.node_repo.save_node_data(updated_node)
                    except Exception as e:
                        logger.error(
                            "Failed to patch node: {}\nresponse: {}", e, resp_text
                        )
                        raise e

        logger.info("Patching node with data: {}", json.dumps(data, indent=4))
        node_id = self.node_repo.get_node_id()
        logger.debug("Node ID: {}", node_id)

        try:
            await __patch_node()
            return Ok("Node patched successfully")
        except aiohttp.ClientResponseError as e:
            # if unauthorized (401) then refresh token and retry
            if e.status == 401:
                logger.debug("Unauthorized, refreshing token ...")
                refresh_result = await self.auth_service.refresh_token()
                if refresh_result.is_err():
                    logger.error(
                        "Failed to refresh token: {}", refresh_result.unwrap_err()
                    )
                    return Err("Failed to refresh token")
                logger.debug("Token refreshed, retrying patch ...")
                await __patch_node()
                return Ok("Node patched successfully")
        except Exception as e:
            err_msg = f"Failed to patch node: {e}"
            logger.error(err_msg)
            return Err(err_msg)

    def _handleDoneTask(self, task: asyncio.Task):
        logger.debug("Handling task completion ...")
        self._tasks.remove(task)
        sys_sweep_task = asyncio.create_task(self._system_sweep_task())
        sys_sweep_task.add_done_callback(self._handleDoneTask)
        self._tasks.add(sys_sweep_task)


async def trigger_system_sweep_hook(plugin_manager: pluggy.PluginManager, node: Node):
    logger.debug("triggering system sweep hook ...")
    results = await plugin_manager.ahook.run_system_sweep(node=node)
    logger.debug("system sweep hook results: {}", results)



import json
from typing import Optional

from ambient_backend_api_client import ContainerRegistryType, EventLabel
from ambient_event_bus_client import Message

from ambient_client_common.utils import logger
from ambient_edge_server.event_handlers.base_handler import BaseHandler
from ambient_edge_server.services.event_service import EventService
from ambient_edge_server.services.registry_service import RegistryServiceFactory


class AuthorizeRegistryHandler(BaseHandler):
    def __init__(
        self, event_svc: EventService, registry_svc_factory: RegistryServiceFactory
    ) -> None:
        super().__init__(event_svc)
        self.registry_svc_factory = registry_svc_factory
        logger.debug("AuthorizeRegistryHandler initialized.")

    @property
    def label(self) -> EventLabel:
        return EventLabel.CONTAINER_AUTH_REQUESTED

    async def handle(self, msg: Message) -> None:
        logger.info("Handling message for topic: {}", msg.topic)
        msg_data: dict = json.loads(msg.message)
        logger.info("Message data: {}", json.dumps(msg_data, indent=4))
        ws_session: Optional[str] = msg_data.get("ws_session", None)
        logger.debug("AuthorizeRegistryHandler.handle() - ws_session: {}", ws_session)

        if ws_session is None:
            logger.error("No websocket session found in message data.")
            return

        registry_type: ContainerRegistryType = msg_data.get(
            "registry_type", ContainerRegistryType.DOCKER_HUB
        )
        registry_svc = self.registry_svc_factory.get_registry_service(registry_type)
        logger.debug(
            "AuthorizeRegistryHandler.handle() - registry_svc: {} - {}",
            registry_svc,
            type(registry_svc),
        )

        logger.info("Authorizing ...")
        logger.debug(
            "AuthorizeRegistryHandler.handle() - authorize method: {}",
            registry_svc.authorize,
        )
        result = await registry_svc.authorize(ws_session=ws_session)
        if result.is_err():
            logger.error("Failed to authorize registry: {}", result.unwrap_err())
            return
        logger.info("Registry authorized successfully.")



# setup.py
from setuptools import setup, find_packages

PACKAGE_NAME = 'ambient_edge_server'
VERSION = '1.2.3'
DESCRIPTION = 'Ambient Edge Server'

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=open('README.md').read(),
    packages=find_packages(),
    install_requires=open('requirements.txt').read().splitlines(),
    entry_points={
        'console_scripts': [
            'ambient_edge_server=ambient_edge_server.run:run',
        ],
    },
    author='Ambient Labs Computing - An Edge Compute Company',
    author_email='jose@ambientlabscomputing.com',
    package_data={'ambient_edge_server': ['templates/*.jinja2', 'assets/*.txt', 'builtin_plugins.yml']},
)



import asyncio
import fnmatch
import importlib
from abc import ABC, abstractmethod
from typing import Dict, Optional, Set, Tuple, Union

import apluggy as pluggy
from ambient_backend_api_client import Configuration as APIConfiguration
from ambient_backend_api_client import NodeOutput as Node
from ambient_base_plugin import BasePlugin
from ambient_base_plugin.models.configuration import ConfigPayload, PluginDefinition
from ambient_event_bus_client import Client, Message, MessageCreate
from fastapi import HTTPException
from result import Err, Result

from ambient_client_common.repositories.docker_repo import DockerRepo
from ambient_client_common.repositories.node_repo import NodeRepo
from ambient_client_common.utils import logger
from ambient_edge_server.config import settings
from ambient_edge_server.repos.plugin_repo import PluginRepo
from ambient_edge_server.services.authorization_service import AuthorizationService

PluginCollection = Dict[str, Tuple[BasePlugin, PluginDefinition]]


class EventService(ABC):
    @abstractmethod
    async def start(self) -> None:
        """Start the event service."""

    @abstractmethod
    async def stop(self) -> None:
        """Stop the event service."""

    @abstractmethod
    async def send_event(self, topic: str, msg: str) -> None:
        """Send an event to the event service.

        Args:
            event (Event): Event to send
        """

    @abstractmethod
    async def handle_plugin_api_request(
        self, method: str, path: str, data: Optional[None] = None
    ):
        """Handle a plugin API request.

        Args:
            method (str): HTTP method
            path (str): URL path
            data (Optional[None], optional): Data to send. Defaults to None.
        """

    @property
    @abstractmethod
    def is_running(self) -> bool:
        """Check if the event service is running.

        Returns:
            bool: True if the event service is running, False otherwise
        """

    @property
    @abstractmethod
    def error(self) -> Union[Result, None]:
        """Get the current error state of the event service.

        Returns:
            Result: The current error state
        """


class AmbientBusEventService(EventService):
    def __init__(
        self,
        client: Client,
        node_repo: NodeRepo,
        plugin_repo: PluginRepo,
        docker_repo: DockerRepo,
        auth_svc: AuthorizationService,
    ) -> None:
        self.client = client
        self.node_repo = node_repo
        self.plugin_repo = plugin_repo
        self.docker_repo = docker_repo
        self.auth_svc = auth_svc
        self.tasks: Set[asyncio.Task] = set()
        self._error = None
        self._is_connected = False
        self.plugins: PluginCollection = {}
        self.plugin_manager = pluggy.PluginManager("ambient_system_sweep")
        self.plugin_manager.add_hookspecs(BasePlugin)

    async def handle_plugin_api_request(
        self, method: str, path: str, data: Optional[None] = None
    ):
        """Handle a plugin API request.

        Args:
            method (str): HTTP method
            path (str): URL path. '/plugins/<plugin_name>/*'
            data (Optional[None], optional): Data to send. Defaults to None.
        """
        path_list = path.split("/")
        path_list = [p for p in path_list if p]
        if len(path_list) < 2:
            logger.error("Invalid path: {}", path)
            return
        plugin_name = path_list[1]
        plugin = find_plugin_from_name(plugin_name, self.plugins)
        if not plugin:
            logger.error("Plugin not found: {}", plugin_name)
            raise HTTPException(
                status_code=404, detail="Plugin not found. Plugin name: {plugin_name}"
            )
        try:
            return await plugin.handle_api_request(method, path, data)
        except HTTPException as e:
            raise e
        except Exception as e:
            logger.error("Error handling API request: {}", e)
            raise HTTPException(
                status_code=500, detail=f"Error handling API request: {e}"
            )

    async def load_plugins(self, api_config: APIConfiguration) -> None:
        logger.info("Loading plugins ...")
        plugins = self.plugin_repo.load_plugins()
        logger.info("Loaded {} plugins.", len(plugins))

        # Register the dynamic hooks with the plugin manager
        logger.info("Registering hooks with pluggy ...")
        for plugin in plugins:
            await self.register_plugin(plugin, api_config)
        logger.info("Plugins loaded.")

    async def register_plugin(
        self, plugin_config: PluginDefinition, api_config: APIConfiguration
    ) -> None:
        """Register a plugin.

        Args:
            plugin_config (PluginDefinition): Plugin Definition
            api_config (APIConfiguration): API configuration
        """
        try:
            logger.debug("Registering plugin: {}", plugin_config.class_name)
            module = importlib.import_module(plugin_config.module)
            logger.debug("Imported module: {}", plugin_config.module)
            plugin_class = getattr(module, plugin_config.class_name)
            logger.debug("Got plugin class: {}", plugin_config.class_name)
            plugin_instance: BasePlugin = plugin_class()
            logger.info("Registered plugin: {}", plugin_config.class_name)
            config_payload = self.build_config_payload(plugin_config, api_config)
            logger.debug(
                "Config payload:\n{}", config_payload.model_dump_json(indent=4)
            )
            await plugin_instance.configure(config=config_payload, logger=logger)
            logger.info("Configured plugin: {}", plugin_config.class_name)

            self.plugins[plugin_config.name] = (plugin_instance, plugin_config)
            self.plugin_manager.register(plugin_instance)
            logger.info("Plugin registered: {}", plugin_config.class_name)
        except ImportError as e:
            logger.error("Could not import module {}: {}", plugin_config.module, e)
        except AttributeError as e:
            logger.error(
                "Attribute error registering plugin {}: {}", plugin_config.class_name, e
            )
            logger.debug(e.with_traceback(e.__traceback__))
        except Exception as e:
            logger.error("Error registering plugin {}: {}", plugin_config.class_name, e)

    def build_config_payload(
        self, plugin_definition: PluginDefinition, api_config: APIConfiguration
    ) -> ConfigPayload:
        logger.info("Building config payload for plugin: {}", plugin_definition.name)
        node_data = self.node_repo.get_node_data()
        config_payload = ConfigPayload(
            node_id=node_data.id,
            plugin_config=plugin_definition,
        )
        if plugin_definition.request_api_config:
            logger.info("Adding API configuration to payload ...")
            config_payload.api_config = api_config
            config_payload.get_token = self.auth_svc.get_token
        if plugin_definition.request_platform:
            logger.info("Adding platform to payload ...")
            config_payload.platform = settings.platform
        if plugin_definition.request_node_repo:
            logger.info("Adding node repo to payload ...")
            config_payload.node_repo = self.node_repo
        if plugin_definition.request_docker_repo:
            logger.info("Adding docker repo to payload ...")
            config_payload.docker_repo = self.docker_repo

        logger.debug("Config payload:\n{}", config_payload.model_dump_json(indent=4))
        return config_payload

    async def start(self, api_config: APIConfiguration) -> None:
        logger.info("Starting event service ...")
        await self.client.init_client()
        node = self.node_repo.get_node_data()
        if node:
            await self.client.add_subscription(f"node-{node.id}/*")
        logger.debug("event client initialized.")

        logger.info("Starting subscription loop ...")
        sub_loop_task = asyncio.create_task(self._subscription_loop())
        sub_loop_task.add_done_callback(self._done_callback)
        sub_loop_task.set_name("subscription_loop_task")
        self.tasks.add(sub_loop_task)
        logger.info("subscription loop started.")

        await self.load_plugins(api_config)

    async def stop(self) -> None:
        logger.info("Stopping event service ...")
        for task in self.tasks:
            logger.debug("cancelling task: {}", task.get_name())
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass

    async def send_event(self, topic: str, msg: str) -> None:
        message_data = MessageCreate(topic=topic, message=msg)
        logger.debug("Publishing message:\n{}", message_data.model_dump_json(indent=4))
        await self.client.publish(msg)

    @property
    def is_running(self) -> bool:
        return all((self._is_connected, not self._error))

    @property
    def error(self) -> Union[Err, None]:
        return self._error

    async def _handle_event(self, msg: Message) -> None:
        logger.debug(f"Handling event for topic: {msg.topic}")

        matched_plugins: Dict[str, Tuple[BasePlugin, PluginDefinition]] = {}
        for plugin_name, plugin in self.plugins.items():
            logger.debug(f"Checking hook: {plugin_name}")
            plugin_instance, plugin_definition = plugin
            for topic in plugin_definition.topics:
                if fnmatch.fnmatch(msg.topic, topic):
                    matched_plugins[plugin_name] = plugin
                    break

        if matched_plugins:
            logger.debug(
                f"Found matching hooks for topic {msg.topic}: {matched_plugins.keys()}"
            )
            for plugin_name, plugin in matched_plugins.items():
                plugin_instance, plugin_definition = plugin
                logger.debug(f"Handling event for plugin: {plugin_name}")
                await plugin_instance.handle_event(msg)
                logger.debug(f"Event handled for plugin: {plugin_name}")
        else:
            logger.warning(f"No matching hooks found for topic: {msg.topic}")

    async def _subscription_loop(self) -> None:
        logger.info("Starting subscription loop ...")
        while True:
            self._error = None
            self._is_connected = True
            try:
                async for message_result in self.client.subscribe():
                    if message_result.is_err():
                        self._error = message_result
                        logger.error(
                            "Error subscribing to messages: {}", message_result
                        )
                        break
                    message = message_result.unwrap()
                    logger.debug(
                        "Received message:\n{}", message.model_dump_json(indent=4)
                    )
                    await self._handle_event(message)
            except Exception as e:
                self._error = Err(e)

    async def _done_callback(self, task: asyncio.Task) -> None:
        err_msg = f"task {task.get_name()} ended. \
Result: {task.result() if task.done() else task.exception()}"
        logger.warning(err_msg)
        self._error = Err(err_msg)

        logger.debug("removing task from tasks set ...")
        self.tasks.remove(task)

        logger.info("restarting subscription loop ...")
        sub_loop_task = asyncio.create_task(self._subscription_loop())
        sub_loop_task.add_done_callback(self._done_callback)
        sub_loop_task.set_name("subscription_loop_task")
        self.tasks.add(sub_loop_task)
        logger.info("subscription loop restarted.")


async def trigger_system_sweep_hook(plugin_manager: pluggy.PluginManager, node: Node):
    logger.debug("triggering system sweep hook ...")
    results = await plugin_manager.ahook.afunc(node=node)
    logger.debug("system sweep hook results: {}", results)


def find_plugin_from_name(
    lowercase_plugin_name: str, plugins: PluginCollection
) -> Union[BasePlugin, None]:
    for plugin_name, plugin in plugins.items():
        if plugin_name.lower() == lowercase_plugin_name:
            return plugin[0]



import json
from typing import Any, Dict

import docker
import websockets
from ambient_backend_api_client import EventLabel, Service
from ambient_event_bus_client import Message
from result import Err, Ok, Result

from ambient_client_common.utils import logger
from ambient_edge_server.event_handlers.base_handler import BaseHandler
from ambient_edge_server.services.event_service import EventService


class DeployServiceHandler(BaseHandler):
    def __init__(
        self, event_service: EventService, docker_client: docker.DockerClient
    ) -> None:
        super().__init__(event_service)
        self.docker_client = docker_client

    async def handle(self, msg: Message) -> None:
        logger.info("Handling deploy service msg: {}", msg.topic)
        logger.debug(msg.model_dump_json(indent=4))

        ws_session_data: Dict[str, Any] = json.loads(msg.message)
        ws_session = ws_session_data.get("ws_session")
        logger.debug("ws_session: {}", ws_session)

        logger.info("connecting to ws session {} ...", ws_session)

        try:
            async with websockets.connect(ws_session) as ws:
                try:
                    logger.info("connected to ws session {}", ws_session)
                    await ws.send("ack")
                    logger.debug("ack sent")

                    service_spec_str = await ws.recv()
                    logger.debug("Received service spec: {}", service_spec_str)

                    service = Service.model_validate_json(service_spec_str)
                    logger.debug(
                        "parsed service spec: {}", service.model_dump_json(indent=4)
                    )

                    logger.info("deploying service {} ...", service.name)
                    result = await self.__deploy_service(service)

                    if result.is_ok():
                        logger.info("service {} deployed", service.name)
                        await ws.send("deployed")
                        return
                    logger.error("error deploying service {}", service.name)
                    await ws.send(f"error: {result.unwrap_err()}")
                except Exception as e:
                    logger.error("Error deploying service: {}", e)
                    await ws.send(f"error: {str(e)}")
        except Exception as e:
            logger.error("Error connecting to ws session: {}", e)

    async def __deploy_service(self, service: Service) -> Result[None, str]:
        """Deploy a service using the docker client

        Args:
            service (Service): Service to deploy

        Returns:
            Result[None, str]: Result of the deployment
        """
        try:
            self.docker_client.services.create(
                image=service.image,
                name=service.name,
                maxreplicas=service.replicas,
                env=service.env_vars,
                mounts=service.mounts,
                networks=service.networks,
                hostname=service.hostname,
            )
            return Ok(None)
        except Exception as e:
            logger.error("Error deploying service {}: {}", service.name, e)
            return Err(str(e))

    @property
    def label(self) -> EventLabel:
        return EventLabel.SERVICE_DEPLOYMENT_REQUESTED



from typing import Optional, Union

import aiohttp
from ambient_backend_api_client import ApiClient, Configuration
from ambient_backend_api_client import NodeOutput as Node
from ambient_backend_api_client import NodesApi, PingApi, TokenResponse
from async_lru import alru_cache
from result import Err, Ok, Result

from ambient_client_common.repositories.node_repo import NodeRepo
from ambient_client_common.utils import logger
from ambient_edge_server.config import settings
from ambient_edge_server.repos.token_repo import TokenRepository


class AuthorizationService:
    def __init__(self, token_repo: TokenRepository, node_repo: NodeRepo) -> None:
        self.token_repo = token_repo
        self.node_repo = node_repo
        self._node_id = None

    @alru_cache(maxsize=1, ttl=3600)
    async def get_token(self) -> Union[str, None]:
        logger.debug("node_id: {}", self.node_id)
        if not self.node_id:
            result = await self.fetch_node()
            logger.debug("result: {}", result)
            if result.is_err():
                logger.error("Error fetching node: {}", result.unwrap_err())
                return None
            self.node_id = result.unwrap().id
        refresh_result = await self.refresh_token()
        if refresh_result.is_err():
            logger.error("Error refreshing token: {}", refresh_result.unwrap_err())
            return None
        return self.token_repo.get_access_token().strip("\n")

    async def verify_authorization_status(self) -> Result[str, str]:
        logger.info("verifying authorization status ...")
        if not self.token_repo.get_access_token():
            logger.error("No access token found in token repo")
            return Err("No access token found in token repo")
        return await self.test_authorization()

    async def fetch_node(self, refresh: bool = False) -> Result[Node, str]:
        if not refresh:
            node = self.node_repo.get_node_data()
            if node:
                logger.debug("node: {}", node.model_dump_json(indent=4))
                return Ok(node)
        node_id = self.node_repo.get_node_id()
        if not node_id:
            return Err("Node not found")
        logger.info("fetching node {} ...", node_id)
        try:
            async with ApiClient(self.api_config) as api_client:
                nodes_api = NodesApi(api_client)
                logger.debug("node_id: {}", node_id)
                if not node_id:
                    return Err("Node ID not found")
                node = await nodes_api.get_node_nodes_node_id_get(node_id=int(node_id))
                logger.debug("node: {}", node.model_dump_json(indent=4))
                self.node_repo.save_node_data(node=node)
                self.node_id = node.id
                return Ok(node)
        except Exception as e:
            err_msg = f"Error fetching node: {e}"
            logger.error(err_msg)
            return Err(err_msg)

    async def authorize_node(self, node_id: str, refresh_token: str) -> None:
        self.node_id = node_id
        async with aiohttp.ClientSession() as session:
            url = settings.backend_api_url
            async with session.post(
                f"{url}/nodes/{node_id}/authorize?refresh_token={refresh_token}"
            ) as response:
                response.raise_for_status()
                token_response = TokenResponse.model_validate(await response.json())
                logger.debug(
                    "token_response: {}", token_response.model_dump_json(indent=4)
                )
                self.token_repo.save_access_token(token_response.access_token)
                self.token_repo.save_refresh_token(token_response.refresh_token)

        await self.get_and_save_node_data(
            node_id=int(node_id), token=token_response.access_token
        )

    async def get_and_save_node_data(
        self, node_id: int, token: str = ""
    ) -> Result[Node, str]:
        logger.info("fetching node {} ...", node_id)
        try:
            async with aiohttp.ClientSession() as session:
                url = settings.backend_api_url
                if not token:
                    token = await self.get_token()
                headers = {"Authorization": f"Bearer {token}"}
                async with session.get(
                    f"{url}/nodes/{node_id}", headers=headers
                ) as response:
                    response.raise_for_status()
                    node = Node.model_validate(await response.json())
                    logger.debug("node: {}", node.model_dump_json(indent=4))
                    self.node_repo.save_node_data(node=node)
                    self.node_id = node.id
                    return Ok(node)
        except Exception as e:
            err_msg = f"Error fetching node: {e}"
            logger.error(err_msg)
            return Err(err_msg)

    async def refresh_token(self) -> Result[None, str]:
        logger.info("refreshing token ...")
        refresh_token = self.token_repo.get_refresh_token()
        if not refresh_token:
            return Err("Refresh token not found")
        logger.debug("refresh_token: {}", refresh_token)
        logger.debug("node_id: {}", self.node_id)
        resp_text: Optional[str] = None
        request_url = f"{settings.backend_api_url}/auth/token?refresh_token\
={refresh_token}"
        logger.debug("request_url: {}", request_url)
        async with aiohttp.ClientSession() as session:
            async with session.post(request_url) as response:
                try:
                    resp_text = await response.text()
                    response.raise_for_status()
                    token_response = TokenResponse.model_validate(await response.json())
                    logger.debug(
                        "token_response: {}", token_response.model_dump_json(indent=4)
                    )
                    self.token_repo.save_access_token(token_response.access_token)
                    self.token_repo.save_refresh_token(token_response.refresh_token)
                    return Ok(None)
                except aiohttp.ClientResponseError as e:
                    logger.error(f"Failed to refresh token: {e}")
                    logger.error(f"response: {resp_text}")
                    return Err(f"Failed to refresh token: {e}")
                except Exception as e:
                    logger.error(f"Failed to refresh token: {e}")
                    logger.error(f"response: {resp_text}")
                    return Err(f"Failed to refresh token: {e}")

    async def create_new_refresh_token(self) -> Result[str, str]:
        logger.info("creating refresh token ...")
        node_data = self.node_repo.get_node_data()
        resp_text: Optional[str] = None
        try:
            headers = {"Authorization": f"Bearer {await self.get_token()}"}
            logger.debug("headers: {}", headers)
            url = settings.backend_api_url
            data = {
                "token_type": "refresh",
                "duration": 3600,
                "user_id": node_data.user_id,
                "org_id": node_data.org_id,
                "node_id": node_data.id,
                "request_type": "node_refresh_token",
            }
            logger.debug("data: {}", data)
            async with aiohttp.ClientSession() as session:
                async with session.post(
                    f"{url}/auth/token-mgmt/", headers=headers, data=data
                ) as response:
                    resp_text = await response.text()
                    logger.debug("response: {}", resp_text)
                    response.raise_for_status()
                    token_response = TokenResponse.model_validate(await response.json())
                    logger.debug(
                        "token_response: {}", token_response.model_dump_json(indent=4)
                    )
                    self.token_repo.save_access_token(token_response.access_token)
                    self.token_repo.save_refresh_token(token_response.refresh_token)
                    return Ok("Refresh token created")
        except Exception as e:
            err_msg = f"Failed to create refresh token: {e}"
            if resp_text:
                err_msg += f"\nresponse: {resp_text}"
            logger.error(err_msg)
            return Err(err_msg)

    async def test_authorization(self) -> Result[str, str]:
        logger.info("testing authorization ...")
        try:
            logger.info("pinging backend ...")
            async with ApiClient(self.api_config) as api_client:
                ping_api = PingApi(api_client)
                pong = await ping_api.auth_ping_auth_ping_get()
                logger.debug("pong: {}", pong)
                logger.info("Authorized")
                return Ok("Authorized")
        except Exception as e:
            logger.error("Error: {}", e)
            return Err("Unauthorized")

    @property
    def node_id(self) -> str:
        return self._node_id

    @node_id.setter
    def node_id(self, value: str) -> None:
        self._node_id = value

    @property
    def api_config(self) -> Configuration:
        return Configuration(
            host=settings.backend_api_url,
            access_token=self.token_repo.get_access_token(),
        )
